#include <jstring.h>
#include <vector>
#include "findcover.h"
#include <jiostr.h>
#include <jfstr.h>
#include <assert.h>


int gcd(int __m, int __n)
{
  while (__n != 0) {
    int __t = __m % __n;
    __m = __n;
    __n = __t;
  }
  return __m;
}

int lcm(int m, int n)
  {
    int g=gcd(m,n);
    return m*n/g;
  }

int main(int argc, char *argv[])
{
  if (argc!=5)
    {
      cerr << "Usage: " << argv[0] << " graphA graphB outputA outputB\n"
              "Input file format:\n"
              "List vertices before edges.  To specify a vertex:\n"
              "   v <name> <block number> <x> <y> <bring forward>\n"
              "Note that the bring forward value is optional.\n"
              "Block numbers should start with 0.\n"
              "To specify an edge:\n"
              "   e <vertex name 1> <vertex name 2> <multiplicity>\n"
              "If no multiplicity is given, 1 is assumed.\n"
              "Output files are suitable for input into drawgraph." << endl;
      return 1;
    }
  Graph graph[2];
  { // read input files
    instl ifstream graphain(argv[1]);
    instl ifstream graphbin(argv[2]);
    if (!graphain || !graphbin)
      {
        cerr << "Error openning file" << endl;
        return 1;
      }
    if (!graph[0].Read(graphain))
      {
        cerr << "Error parsing file " << argv[1] << endl;
        return 1;
      }
    if (!graph[1].Read(graphbin))
      {
        cerr << "Error parsing file " << argv[2] << endl;
        return 1;
      }
  }

  // compute MaxRType
  if (graph[0].mMaxRType!=graph[1].mMaxRType)
    {
      cerr << "Incompatible number of refined types in graphs\n";
      return 1;
    }
  int MaxRType=graph[0].mMaxRType;

  // compute g()
  Gfunct g[2];
  g[0].Initialize(&graph[0], MaxRType);
  g[1].Initialize(&graph[1], MaxRType);
  int i, j;

  // compute m_{ij}
  instl vector<int> gM(MaxRType*MaxRType, 0);
  for (i=0; i<graph[0].mE.size(); ++i)
    {
      int vh=graph[0].mE[i].mV[0],
          vt=graph[0].mE[i].mV[1];
      int typeh=graph[0].mV[vh].refinedtype,
          typet=graph[0].mV[vt].refinedtype;
      gM[typeh*MaxRType + typet] += 1;
      gM[typet*MaxRType + typeh] += 1;
    }

  // compute n_i
  instl vector<int> gN(MaxRType);
  for (i=0; i<graph[0].mV.size(); ++i)
    ++gN[graph[0].mV[i].refinedtype];

  // compute s
  int s=0;
  for (i=0; i<MaxRType; ++i)
    for (j=i; j<MaxRType; ++j)
      {
        int mv=gM[i*MaxRType+j];
        if (mv!=0)
          if (s!=0)
            s=lcm(mv,s);
          else
            s=mv;
      }

  // compute a_i
  instl vector<int> gA(MaxRType);
  for (i=0; i<MaxRType; ++i)
    if (gN[i])
      {
        assert(s % gN[i] == 0);
        gA[i]=s/gN[i];
      } 
    else
      gA[i]=0;

  // compute b_{ij}
  instl vector<int> gB(MaxRType*MaxRType);
  for (i=0; i<MaxRType*MaxRType; ++i)
    if (gM[i])
      {
        assert(s % gM[i] == 0);
        gB[i]=s/gM[i];
      } 
    else
      gB[i]=0;  // need not be defined


  // define graph out
  Graph out;

  // add vertices to out
  instl vector<int> gAsummed(graph[0].mV.size());
  {
    int sum=0;
    i=0;
    for (Graph::VertexIter v=graph[0].mV.begin();
         v!=graph[0].mV.end();
         ++v)
      {
        gAsummed[i]=sum; // i is getting too big
        ++i;
        for (Graph::VertexIter vp=graph[1].mV.begin();
             vp!=graph[1].mV.end();
             ++vp)
          if ((*v).refinedtype == (*vp).refinedtype)
            {
              int ai=gA[(*v).refinedtype];
              sum+=ai;
              for (int p=0; p<ai; ++p)
                {
                  Vertex vh;
#if defined(__GNUC__)
                  int dec=0, sign=0;
                  char *pstr=fcvt(p, 0, &dec, &sign );
#else
                  static char pstr[12];
                  itoa(p, pstr, 10);
#endif
                  vh.mName=(*v).mName+"_"+(*vp).mName+"_"+pstr;
                  vh.degree=(*v).degree;
                  if ((*v).degree != (*vp).degree)
                    {
                      cerr << "Vertices of same type have different degrees\n";
                      return 1;
                    }
                  vh.refinedtype=(*v).refinedtype;
                  out.mV.push_back(vh);
                }
            }
      }
  }

  // compute edges for out
//  cerr << graph[0].mE.size() << "?\n";
  {
    for (int e=0; e<graph[0].mE.size(); ++e)
      for (int ep=0; ep<graph[1].mE.size(); ++ep)
        for (int erev=0; erev<=Reversed; erev+=Reversed)
          for (int eprev=0; eprev<=Reversed; eprev+=Reversed)
            {
//              cerr << e << ' ' << erev << ' ' << ep << ' ' << eprev << '\n';
              int vt=graph[0].Tail(e+erev);
              int vtp=graph[1].Tail(ep+eprev);
              i=graph[0].RType(vt);
              if (i!=graph[1].RType(vtp))
                continue;
              int vh=graph[0].Head(e+erev);
              int vhp=graph[1].Head(ep+eprev);
              j=graph[0].RType(vh);
              if (j!=graph[1].RType(vhp))
                continue;
              int ij=i*MaxRType+j;
              int ji=j*MaxRType+i;
              for (int q=0; q<gB[ij]; ++q)
                {
                  int rij=g[0].mR[ij];
                  int V[2];
                  // tail
                  int gdiff = g[0](vt, e+erev)-g[1](vtp, ep+eprev);
                  gdiff %= rij;
                  if (gdiff<0) 
                    gdiff+=rij;
                  int p = rij * q + gdiff;
                  int ai=gA[i];
                  //V[0]=p+gA[i]*vtp+gAsummed[vt];
                  V[0]=p+gAsummed[vt];
                  int zz;
                  for (zz=0; zz<vtp; ++zz)
                    //if (vertex zz in graph[1] is of type i)
                    if (graph[1].RType(zz)==i)
                      V[0]+=ai;
#ifndef NDEBUG                  
#if defined(__GNUC__)
                  int dec=0, sign=0;
                  char *pstr=fcvt(p, 0, &dec, &sign );
#else
                  static char pstr[12];
                  itoa(p, pstr, 10);
#endif
                  string name=graph[0].mV[vt].mName+"_"+graph[1].mV[vtp].mName+"_"+pstr;
                  if (out.mV[V[0]].mName!=name)
                    {
                      cerr << V[0] << ' ' << p << ' ' << i << ' ' << gA[i] << ' ' << vtp << ' ' 
                           << vt << ' ' << gAsummed[vt] << ' ' << graph[1].mV.size() << endl;
                      V[0]=out.FindVert(name);
                      assert(V[0]>=0);
                      assert(0);
                    }
#endif

                  // head
                  rij = g[0].mR[ji];
                  gdiff = g[0](vh, e+Reversed-erev)-g[1](vhp, ep+Reversed-eprev);
                  gdiff %= rij;
                  if (gdiff<0) 
                    gdiff+=rij;
                  p = rij * q + gdiff;
                  ai=gA[j];
                  //V[1]=p+gA[j]*vhp+gAsummed[vh];
                  V[1]=p+gAsummed[vh];
                  for (zz=0; zz<vhp; ++zz)
                    //if (vertex zz in graph[1] is of type j)
                    if (graph[1].RType(zz)==j)
                      V[1]+=ai;

#ifndef NDEBUG                  
#if defined(__GNUC__)
                  pstr=fcvt(p, 0, &dec, &sign );
#else
                  itoa(p, pstr, 10);
#endif
                  name=graph[0].mV[vh].mName+"_"+graph[1].mV[vhp].mName+"_"+pstr;
                  if (out.mV[V[1]].mName!=name)
                    {
                      //assert(0);
                      cerr << V[1] << ' ' << p << ' ' << j << ' ' << gA[j] << ' ' << vhp << ' ' 
                           << vh << ' ' << gAsummed[vh] << ' ' << graph[1].mV.size() << endl;
                      V[1]=out.FindVert(name);
                      assert(V[1]>=0);
                      assert(0);
                    }
#endif

                  // check if present but undoubled
                  int oe;
                  for (oe=0; oe<out.mE.size(); ++oe)
                    if (out.mE[oe].mDoubled==false &&
                        out.mE[oe].mV[1]==V[0] &&
                        out.mE[oe].mV[0]==V[1])
                      break;
                  if (oe<out.mE.size())
                    // present, undoubled, so double
                    out.mE[oe].mDoubled=true;
                  else
                    // otherwise:
                    out.mE.push_back(Edge(V));
                }
            }
  }

  for (i=0; i<2; ++i)
    {
      instl ofstream outputfile(argv[3+i]);
      outputfile << "// Graph " << i << ": " << argv[i+1] << '\n';
      graph[i].Output(outputfile);
      outputfile << '\n';
      outputfile << "// Cover of " << argv[1] << " and " << argv[2] << "\n";
      outputfile << "c\n";
      out.CoverOutput(outputfile);
    }
  return 0;
}

